/*
 * Copyright (C) 2002  Robert Siemer <Robert.Siemer-gdoamin@backsla.sh>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */
#define _GNU_SOURCE

#define DEBUG 1

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <search.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "config2.h"


typedef void (*vf)(void *);

pthread_mutex_t debug_lock = PTHREAD_MUTEX_INITIALIZER;

#if DEBUG
#define d(call) do { \
	pthread_cleanup_push((vf)pthread_mutex_unlock, (void *) &debug_lock); \
	pthread_mutex_lock(&debug_lock);                                      \
	call;                                                                 \
	pthread_cleanup_pop(1);                                               \
} while (0)
#else
#define d(call)
#endif

struct order {
	char *command;
	char *target;
	struct order *next;
};

struct handle {
	int key; /* file descriptor */
	char buffer[INPUT_BUFFER];
	size_t offset;
	struct order *order;
};

/* Order queues. Higher number have lower priorities. */

struct order* volatile queue[NR_QUEUES];
struct order*          queue_end[NR_QUEUES];

/* One 'read' lock for both queues. ('read' is symbolic. The lock is for
   reading the queues, but you can modify (r/w) them. */

pthread_mutex_t read_lock = PTHREAD_MUTEX_INITIALIZER;

/* Condition if one of the queues goes from empty to nonempty. Maybe just
   used if all queues were empty (not just one). */

pthread_cond_t queue_nonempty = PTHREAD_COND_INITIALIZER;

pthread_t worker[WORK_INSTANCES];

void *handle_tree;


void error_exit(char *string) __attribute__ ((noreturn));
void error_exit(char *string)
{
	#define ER_EX_SIZE 100
	char combo[ER_EX_SIZE];
	snprintf(combo, sizeof(combo), "%s %s",
		 program_invocation_short_name, string);
	combo[sizeof(combo)-1]=0;
	perror(combo);

	exit(EXIT_FAILURE);
}

#define RETRY_IF(loopcondition, error)                  \
do {                                                    \
	int counter=RETRY_COUNT;                        \
	while ((loopcondition) && counter) {            \
		d(puts("LOOPING!!"));                   \
		perror(program_invocation_short_name);  \
		counter--;                              \
		sleep(RETRY_SLEEP);                     \
	}                                               \
	if (!counter) error_exit(error);                \
} while (0);

/* return next order; can block! */
struct order* getorder()
{
	struct order* ret;
	int prio;

	d(puts("Entering getorder()"));

	pthread_cleanup_push((vf)pthread_mutex_unlock, (void *) &read_lock);
	pthread_mutex_lock(&read_lock);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
 start:
	for (prio=0; prio<NR_QUEUES; prio++) {
		d(printf("Checking prio %d, which is %d.\n", prio,
			 (int) queue[prio]));
		if (queue[prio]) {
			ret = queue[prio];
			queue[prio]=queue[prio]->next;
			goto end;
		}
	}
	d(puts("Entering cond_wait()"));
	pthread_cond_wait(&queue_nonempty, &read_lock);
	d(puts("Leaving cond_wait()"));
	goto start;
 end:
	pthread_cleanup_pop(1);
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

	d(printf("Returning order: %s, %s\n", ret->command,
		 ret->target));

	return ret;
}

/* just one thread will/is allowed to call putorder() - it's not reentrant */
void putorder (struct order* order, int prio)
{
	order->next=NULL;
	if (prio >= NR_QUEUES) prio = NR_QUEUES-1;

	d(printf("Order: %s, %s; prio: %d\n", order->command, order->target,
		 prio));

	pthread_mutex_lock(&read_lock);
	if(queue[prio]) {  /* queue not empty; end is valid */
		d(puts("Non-empty."));
		queue_end[prio]->next=order;
	} else {           /* queue is empty; end is invalid */
		queue[prio]=order;
		pthread_cond_broadcast(&queue_nonempty);
	}
	d(printf("Queue starts at: %d for %d.\n", (int)queue[prio], prio));
	pthread_mutex_unlock(&read_lock);
	queue_end[prio]=order;
}

void destroy_order(struct order* order)
{
	free(order->command);
	free(order->target);
	free(order);
}

/* return a fd of a temporary file and let the filename be in temp_filename;
   malloc space for that
*/
int tmp_open(char **temp_filename, const char *target)
{
	int ret;

	puts("Entering tmp_open()");
	RETRY_IF(asprintf(temp_filename, "%s.tmp.XXXXXX", target)<0,
		"asprintf()");
	printf("mkstemp(%s)...\n", *temp_filename);
	RETRY_IF((ret=mkstemp(*temp_filename))<0, "mkstemp()");
	puts("Leaving tmp_open()");
	return ret;
}

void *work_thread(void *pt)
{
	struct order *order;
	int target_fd, tmp_fd, pid;
	char *tmp_file;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

 loop:
	pthread_testcancel();
	order = getorder();
	d(printf("Opening target %s.\n", order->target));
	target_fd = open(order->target, O_CREAT|O_EXCL|O_WRONLY, TARGET_FILE_MODE);
	if (target_fd >= 0) {
		d(puts("Attend that work..."));
		close(target_fd);
		tmp_fd=tmp_open(&tmp_file, order->target);
		d(puts("Before fork()"));
		RETRY_IF((pid=fork())<0, "fork()");
		d(printf("After fork(): PID IS %d!\n", pid));
		if (pid == 0) { /* child executes 'command' */
			dup2(tmp_fd, STDOUT_FILENO);
			close(tmp_fd);
			RETRY_IF(execl(SHELL, SHELL, "-c", order->command,
				NULL)<0, "execl()");
		} else { /* parent cleans up, waits and cleans up */
			close(tmp_fd);
			chmod(tmp_file, TARGET_FILE_MODE);
			waitpid(pid, NULL, 0);
			d(printf("Renameing %s to %s.\n", tmp_file,
					order->target));
			if(rename(tmp_file, order->target)==-1)
				error_exit("rename()");
			free(tmp_file);
			d(puts("Cleaning up..."));
			destroy_order(order);
		}
	} else {
		d(puts("Skip that work..."));
		if (errno != EEXIST)
			perror(order->target);
		destroy_order(order);
	}
	goto loop;

	return NULL;
}

int handle_comp (const void *one, const void *two)
{
	int key1 = ((struct handle*) one)->key;
	int key2 = ((struct handle*) two)->key;
	
	return key1-key2;
}

void create_n_add_handle(int fd)
{
	struct handle* handle;
	struct handle** h2;

	printf("fd: %d\n", fd);

	/* create and init handle without order */
	RETRY_IF((handle=malloc(sizeof(struct handle)))==NULL, "malloc()");
	handle->order = NULL;
	handle->key = fd;
	handle->offset = 0;

	/* add handle to tree */
	errno=0;
	RETRY_IF((h2=tsearch(handle,&handle_tree,handle_comp))==0, "tsearch()");
	if (*h2 != handle) error_exit("tsearch");
}

void remove_n_destroy_handle(struct handle *hp)
{
	if (!tdelete(hp, &handle_tree, handle_comp)) error_exit("tdelete");
	if (hp->order) {
		if (hp->order->target) free(hp->order->target);
		if (hp->order->command) free(hp->order->command);
		free(hp->order);
	}
	free(hp);
}

void handle2order(struct handle *handle)
{
	char *newline, *string;
	int size, prio;

	/* We don't read something here. We work with the data in ->buffer */
 loop:
	newline = memchr(handle->buffer, '\n', handle->offset);
	if (!newline) return;
	size = newline - handle-> buffer;
	RETRY_IF((string=malloc(size+1))==NULL, "malloc()");
	memcpy(string, handle->buffer, size);
	string[size]=0;
	if (!handle->order) { /* I got the command. */
		RETRY_IF((handle->order=malloc(sizeof(handle->order)))==NULL,
			"malloc()");
		handle->order->command = string;
	} else {  /* I got the target. */
		handle->order->target = string;
		prio=strspn(handle->order->command, " ");
		putorder(handle->order, prio);
		handle->order = NULL;
	}
	/* 'clean up' ->buffer */
	memmove(handle->buffer, handle->buffer+size+1,
		handle->offset - (size+1));
	handle->offset -= size+1;
	goto loop;
}

/* return non-null if connection closed or is to close */
int get_handle_n_handle(int fd)
{
	struct handle h;
	struct handle *handle;
	struct handle **h2;
	ssize_t i;

	h.key=fd;
	h2 = tfind(&h, &handle_tree, handle_comp);
	if (!h2) error_exit("tfind");
	handle = *h2;
	if (!handle) error_exit("tfind 2");

 loop:
	i=read(fd, handle->buffer+handle->offset,
	       sizeof(handle->buffer)-handle->offset);
	if (i > 0) {
		handle->offset += i;
		handle2order(handle);
		goto loop;
	}
	if (i ==0) {
		remove_n_destroy_handle(handle);
		return -1;
	} else { /* (i<0) */
		switch (errno) {
		case EAGAIN:
			return 0;
		default:
			error_exit("read");
		}
	}
}

void nonblock(int fd)
{
	int flags = fcntl (fd, F_GETFL, 0);
	if (flags == -1) error_exit("fcntl(F_GETFL...)");
	flags |= O_NONBLOCK;
	flags = fcntl (fd, F_SETFL, flags);
	if (flags == -1) error_exit("fcntl(F_SETFL...)");
}

void server (void)
{
	int e, i, sock, nfds;
	struct sockaddr_un name;
	fd_set active_fd_set, read_fd_set;
	
	sock = socket(PF_LOCAL, SOCK_STREAM, 0);
	if (sock < 0) error_exit("socket");
	nonblock(sock);
	i=1;
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(int)))
		error_exit("setsockopt");

	if (sizeof(SOCKET_FILENAME) > sizeof(name.sun_path))
		error_exit("SOCKET_FILENAME too long...");
	unlink(SOCKET_FILENAME);
	strncpy(name.sun_path, SOCKET_FILENAME, sizeof(name.sun_path));
	name.sun_family=AF_LOCAL;
	e = bind(sock, (struct sockaddr*) &name, SUN_LEN(&name));
	if (e) error_exit("bind");
	e = listen(sock, 16);
	if (e) error_exit("listen");
	
	/* Initialize the set of active sockets. */
	FD_ZERO (&active_fd_set);
	FD_SET (sock, &active_fd_set);
	nfds = sock+1;

 loop:
	/* Block until input arrives on one or more active sockets. */
	read_fd_set = active_fd_set;
	e = select (nfds, &read_fd_set, NULL, NULL, NULL);
	if (e < 0) error_exit ("select");

	/* Service all the sockets with input pending. */
	for (i = 0; i < nfds; ++i)
		if (FD_ISSET (i, &read_fd_set))	{
			if (i == sock) {
				int new;
				new = accept (sock, NULL, NULL);
				if (new < 0) error_exit("accept");
				nonblock(new);
				FD_SET (new, &active_fd_set);
				create_n_add_handle(new);
				if (new > nfds-1)
					nfds=new+1;
			} else {
				/* old socket */
				if (get_handle_n_handle(i)) {
					close (i);
					FD_CLR (i, &active_fd_set);
				}
			}
		}
	goto loop;
}

void worker_on(pthread_t *thread)
{
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED))
		error_exit("pthread_attr_setdetachstate");
	if (pthread_create(thread, &attr, work_thread, NULL))
		error_exit("pthread_create");
	pthread_attr_destroy(&attr);
}

#define worker_off(thread) pthread_cancel(thread)

sigset_t block_signals(void)
{
  sigset_t int_signals, orig_signals;

  sigemptyset (&int_signals);
  sigaddset(&int_signals, SIGINT);
  sigaddset(&int_signals, SIGTERM);

  sigprocmask(SIG_BLOCK, &int_signals, &orig_signals);

  return orig_signals;
}

void int_handler (int sig) {
	int i;

	for (i=0; i<WORK_INSTANCES; i++) {
		worker_off(worker[i]);
	}

	unlink(SOCKET_FILENAME);
	pthread_exit(NULL);
}

void register_and_unblock(sigset_t orig_signals)
{
  signal(SIGINT, int_handler);
  signal(SIGTERM, int_handler);

  sigprocmask(SIG_SETMASK, &orig_signals, 0);
}

int main (int argc, char *argv[])
{
	int pid, i;
	sigset_t orig_signals;
	
	orig_signals = block_signals();

	for (i=0; i<WORK_INSTANCES; i++) {
		worker_on(&worker[i]);
	}
	register_and_unblock(orig_signals);
	server();

	return 0;
}

